import os
import zipfile
import csv
import shutil
import tempfile
from collections import namedtuple
from operator import itemgetter


#todo
#pylint?
#remove? duplicated geocheem data table (waste of mem?)
class GasFileLoader:

    Column = namedtuple('Column', 'name type index')

    def intVer(ver):
        try:
            vers = ver.split(".")
            ret = int(vers[0]) * 1000 + int(vers[1])

        except:
            ret = 999999
        #print(f"ver = {ver} ret = {ret}")
        return ret


    def __init__(self, gasfile):

        # UI and version strings
        ERR_EN = "East and North fields must be set in ioGAS prior to saving"
        REQUIRED_VERSION = "6.3"

        self.gasfile = gasfile
        ### Public fields (the non list declarations here are for doco only!)

        #version of loaded gas file as int  - eg 6.3.1 is 6003; 7.0 is 7000
        self.version = 0
        self.versionString = ""

        #lists of attributes as named tuples
        self.filters = []
        self.colours = []
        self.shapes = []
        self.sizes = []

        #att titles
        self.colourTitle = ""
        self.shapeTitle = ""
        self.sizeTitle = ""
        self.symbologyTitle = ""

        # symbology field names
        self.GAS_SIZE = "ioGAS_Size__"
        self.GAS_SHAPE = "ioGAS_Shape__"
        self.GAS_COLOUR = "ioGAS_Colour__"
        self.GAS_FILTER = "ioGAS_Filter__"

        #list of selected Column named tuples from xml
        self.columns = []

        #list of rows as tuples of cells
        self.data = [] #todo get rid of this?

        #list of rows attribute indices as 4-tuples of indices
        self.atts = []

        #both...
        self.dataAndAtts = []

        #list of unique symbology combinations as sorted list of unique of 3-tuples of indices
        self.uniqueAttCombinations = []

        #special cols
        self.specialColumnEast = ""
        self.specialColumnNorth = ""
        self.specialColumnElevation = ""
        self.specialColumnID = ""

        # column index of coord fields
        self.indexX = -1
        self.indexY = -1
        self.indexZ = -1

        #proj
        self.epsg = ""

        ### Impl

        #unzip
        tempDir = tempfile.mkdtemp() #cleaned up at end
        zip_ref = zipfile.ZipFile(gasfile, 'r')
        zip_ref.extractall(tempDir)
        zip_ref.close()
        dataFile = os.path.join(tempDir ,"data.csv")
        metadataFile = os.path.join(tempDir ,"metadata.xml")
        versionFile = os.path.join(tempDir, "version.txt")

        ### 0. read version, and check it
        f = open(versionFile)
        v = list(csv.reader(f))[0][0] #single cell - the version like 6.3.1
        ver = GasFileLoader.intVer(v)
        self.version = ver
        self.versionString = v
        f.close()
        if self.version < GasFileLoader.intVer(REQUIRED_VERSION):
            raise ValueError(f"File must have been saved with ioGAS version {REQUIRED_VERSION} or later")



        ### 1. read metadata
        import xml.etree.ElementTree as ET
        doc = ET.parse(metadataFile).getroot() #nb encoding is set in the file, so no need to reset

        Filter = namedtuple('Filter', 'name visible')
        fas = doc.find("extraAttributes2")
        for fa in fas.findall("extraAttribute"):
            filter = Filter(GasFileLoader.unescape(fa.find("name").text),
                            fa.find("visible").text == 'true')
            self.filters.append(filter)

        Colour = namedtuple('Colour', 'name colour visible')
        cas = doc.find("colourAttributes")
        self.colourTitle = GasFileLoader.unescape(cas.find("title").text).strip()
        for ca in cas.findall("colourAttribute"):
            colour = Colour(GasFileLoader.unescape(ca.find("name").text),
                            ca.find("colour").text,
                            ca.find("visible").text == 'true')
            self.colours.append(colour)

        Shape = namedtuple('Shape', 'name shapeCode visible')
        sas = doc.find("shapeAttributes")
        self.shapeTitle = GasFileLoader.unescape(sas.find("title").text).strip()
        for sa in sas.findall("shapeAttribute"):
            shape = Shape(GasFileLoader.unescape(sa.find("name").text),
                          sa.find("shapeCode").text,
                          sa.find("visible").text == 'true')
            self.shapes.append(shape)

        Size = namedtuple('Size', 'name size visible')
        zas = doc.find("sizeAttributes")
        self.sizeTitle = GasFileLoader.unescape(zas.find("title").text).strip()
        for za in zas.findall("sizeAttribute"):
            size = Size(GasFileLoader.unescape(za.find("name").text),
                        za.find("size").text,
                        za.find("visible").text == 'true')
            self.sizes.append(size)


        self.symbologyTitle = self.colourTitle + ", " + self.shapeTitle + ", " + self.sizeTitle

        # all columns into a dict (so we can get properties below)

        columns = {}
        cols = doc.find("columns")
        index=0
        for col in cols.findall("column"):
            name = GasFileLoader.unescape(col.find("aliasName").text)
            c = GasFileLoader.Column(name, col.find("type").text, index)
            columns[name] = c
            index = index +1

        # optional metadata
        sc = doc.find("specialColumns")
        # nb special cols always exported, even if not selected
        try:
            self.specialColumnID = GasFileLoader.unescape(sc.find("id").text)
            GasFileLoader.safeAppend(self.columns, (columns[self.specialColumnID]))
        except:
            pass


        try:
            self.specialColumnEast = GasFileLoader.unescape(sc.find("map_east").text)
            GasFileLoader.safeAppend(self.columns, columns[self.specialColumnEast]) #always export specials
            #print(f"!!!outer list is {self.columns}")
            self.indexX = [i for i, c in enumerate(self.columns) if (c.name == self.specialColumnEast)][0]
        except:
            raise (ValueError(ERR_EN))

        try:
            self.specialColumnNorth = GasFileLoader.unescape(sc.find("map_north").text)
            GasFileLoader.safeAppend(self.columns, columns[self.specialColumnNorth])  # always export specials
            self.indexY = [i for i, c in enumerate(self.columns) if (c.name == self.specialColumnNorth)][0]
        except:
            raise (ValueError(ERR_EN))

        try:
            self.specialColumnElevation = GasFileLoader.unescape(sc.find("map_elevation").text)
            GasFileLoader.safeAppend(self.columns, columns[self.specialColumnElevation])  # always export specials
            if self.specialColumnElevation != None and self.specialColumnElevation:
                self.indexZ = [i for i, c in enumerate(self.columns) if (c.name == self.specialColumnElevation)][0]
            else:
                self.indexZ = -1
        except:
            self.indexZ = -1   # valid to have no elevation column, but we flag as missing

        try:
            self.epsg = GasFileLoader.unescape(sc.find("map_epsg").text)
        except:
            pass

        #only add the selected columns, in order
        selCols = []
        vshist = doc.find("variableSelectionHistory")
        vsh = vshist.findall("variableSelection")
        if len(vsh) > 0:
            vs = vsh[-1] #last
            for sel in vs.findall("variable"):
                name = GasFileLoader.unescape(sel.find("variableName").text)
                column = columns[name]
                if (column not in self.columns):
                    # prevent adding special cols twice because we only store a single index of col pos
                    self.columns.append(column)

        ### 2.  read data and atts from csv


        #nb the commented out line here breaks in both QGIS 3.2 and 3.4
        #f = open(dataFile, encoding='utf-8') #utf-8 is default in QGIS 3.4 (but not 3.2)
        f = open(dataFile, encoding='Cp1252') #.gas files use THIS encoding for the internal CSV (as its still the default on windows?)

        headerRow = True
        for rawLine in csv.reader(f):
            line = (GasFileLoader.unescape(cell) for cell in rawLine)
            tup = tuple(line)
            data = GasFileLoader.lookup(tup, self.columns)
            att = tup[-4:]
            dataAndAtt = data + att
            if headerRow:
                pass
                # unused self.columnNames = data
            else:
                #IF row is visible
                if self.filters[int(att[0])].visible \
                    and self.colours[int(att[1])].visible \
                        and self.shapes[int(att[2])].visible \
                            and self.sizes[int(att[3])].visible:
                                #print("col ", allColours[int(att[1])], "tup = ", tup, "  adding ROW", line) ####
                                self.data.append(data)
                                self.atts.append(att)
                                self.dataAndAtts.append(dataAndAtt)
                else:
                    pass
                    #print("SKIPPING ROW", line)
            headerRow = False
        f.close()

        # build unique att combinations, and sort like gas
        uac = set()
        for row in self.atts:
            intRow = tuple(int(i) for i in row)
            uac.add(intRow)
        self.uniqueAttCombinations = sorted(uac, key=itemgetter(1, 2, 3))

        #clean up
        shutil.rmtree(tempDir)

    #lookup data value by index (todo slow?)
    def lookup(line, cols):
        ret = []
        for col in cols:
            index = col.index
            value = line[index]
            ret.append(value)
        return tuple(ret)


    #gas escaping rules
    def unescape(cell):
        ret =  cell.replace("_a", "&")
        ret = ret.replace("_l", "<")
        ret = ret.replace("_g", ">")
        ret = ret.replace("_q", "\"")
        ret = ret.replace("_c", ",")
        ret = ret.replace("_t", "\t")
        ret = ret.replace("_u", "_")
        return ret

    # appends, changing name if needed to enforce uniqueness
    def safeAppend(cols, col):
        n = col.name
        while (GasFileLoader.containsName(cols, n)):
            n = n + "1"
        col = GasFileLoader.Column(n, col.type, col.index)
        cols.append(col)


    # already have a column of this name?
    def containsName(cols, colName):
        for c in cols:
            if c.name == colName:
                return True
        return False







