

import os
from qgis.PyQt.QtCore import QCoreApplication, QTranslator
from qgis.PyQt.QtGui import *
from qgis.PyQt.QtCore import QFileSystemWatcher
from qgis.PyQt.QtWidgets import QAction
from qgis.core import * # QgsApplication, QgsVectorLayer, QgsProject, QgsMessageLog, QgsSettings
from pathlib import Path
from PyQt5.QtWidgets import QFileDialog, QMenu,QToolButton

#start of this path 'ioGAS' must match deployment
from ioGAS.gasfileloader import GasFileLoader
from ioGAS.gaslogic import GasLogic

import tempfile

pluginPath = os.path.dirname(__file__)

buildID  = "104612"

LIMIT = 60  # max number of unique attributes we can handle in Lnear legend mode on qgis < 3.3 (sqllite limitation)

vectorMenuText ='Import ioGAS File'

class QGisPluginGas:
    def __init__(self, iface):
        self.iface = iface
        self.legendType = 'LINEAR' # Hierachical / Linear; currently we are only supporting Linear
        self.layerType = '' # Scratch / Geopackage

    # set this plugin up in QGIS UI : toolbar dropdown and vector menu submenu
    def initGui(self):

        # Create 2 actions - for scracth and geopackage

        self.actionA = QAction('Scratch Layer', self.iface.mainWindow())
        self.actionA.setIcon(QIcon(os.path.join(pluginPath, 'icons', 'iogas.svg')))
        self.actionA.triggered.connect(self.runScratch)

        self.actionC = QAction('Create GeoPackage', self.iface.mainWindow())
        self.actionC.setIcon(QIcon(os.path.join(pluginPath, 'icons', 'iogas.svg')))
        self.actionC.triggered.connect(self.runLinear)

        #self.actionB = QAction('With Tree Legend', self.iface.mainWindow())
        #self.actionB.setIcon(QIcon(os.path.join(pluginPath, 'icons', 'iogasTree.svg')))
        #self.actionB.triggered.connect(self.runHierarchical)



        # Add the actions as submenu items to the Vector menu
        self.iface.addPluginToVectorMenu(vectorMenuText, self.actionA)   # submenu name, menuItem
        #self.iface.addPluginToVectorMenu(vectorMenuText, self.actionB)
        self.iface.addPluginToVectorMenu(vectorMenuText, self.actionC)

        # For the toolbar widget we create  a popup menu triggered from a toolbutton
        self.popupMenu = QMenu(self.iface.mainWindow())
        self.popupMenu.addAction(self.actionA)
        #self.popupMenu.addAction(self.actionB)
        self.popupMenu.addAction(self.actionC)

        self.toolButton = QToolButton()
        self.toolButton.setIcon(QIcon(os.path.join(pluginPath, 'icons', 'iogas.svg')))
        self.toolButton.setToolTip(vectorMenuText)
        self.toolButton.setMenu(self.popupMenu)
        self.toolButton.setPopupMode(QToolButton.InstantPopup)

        self.actionToolButton = self.iface.addToolBarWidget(self.toolButton)

        self.taskManager = QgsApplication.taskManager()

        QgsMessageLog.logMessage(f"Loaded QGIS for ioGAS Plugin V {buildID}" ,  tag="ioGAS", level=Qgis.Info)


    def unload(self):
        self.iface.removePluginVectorMenu(vectorMenuText, self.actionA)  # submenu name, menuItem
        #self.iface.removePluginVectorMenu(vectorMenuText, self.actionB)  # submenu name, menuItem
        self.iface.removePluginVectorMenu(vectorMenuText, self.actionC)  # submenu name, menuItem

        self.iface.removeToolBarIcon(self.actionToolButton )


    # entry points
    def runLinear(self):
        self.legendType = 'LINEAR'
        self.layerType = 'GEOPACKAGE'
        self.run()
        return

    # def runHierarchical(self):
    #     self.legendType = 'HIERARCHICAL'
    #     self.layerType = 'GEOPACKAGE'
    #     self.run()
    #     return

    def runScratch(self):
        self.legendType = 'LINEAR'  # todo resolve user choice here
        self.layerType = 'SCRATCH'
        self.run()
        return


    def run(self):
            #input (prob win reg - yuck)
            path = str(QgsSettings().value('ioGAS-Qgis/lastInputDirectory', Path.home()))
            QgsMessageLog.logMessage(f"got last path {path}", 'ioGAS', level=Qgis.Info)
            chosen = QFileDialog.getOpenFileName(self.iface.mainWindow(), "Open ioGAS file", path, "*.gas")
            src = chosen[0]
            if src=='':
                return #cancel
            lastDir = os.path.dirname(src)
            QgsMessageLog.logMessage(f"saving last path {lastDir}", 'ioGAS', level=Qgis.Info)
            QgsSettings().setValue('ioGAS-Qgis/lastInputDirectory', lastDir)

            layerName = os.path.splitext(os.path.basename(src))[0]  # filename only w/out extension



           # Read the Gas file
            app = GasLogic()
            app.setLegendType(self.legendType)
            reader = GasFileLoader(src)
            QgsMessageLog.logMessage(f"loading {src} version {reader.versionString} ", 'ioGAS', level=Qgis.Info)

            app.setReader(reader)


            #check for legend limitation (Qgis Ver vs Max number of unique combinations for linera legends)
            if not app.canDisplayLegend():
                msg = f'Cannot display this file - must be fewer than {app.LEGEND_LIMIT} unique attributes for Linear Legend'
                QgsMessageLog.logMessage(msg,'ioGAS', level=Qgis.Warning)
                self.iface.messageBar().pushWarning('ioGAS',msg)
                return

            # Option 1: Create a Geopackage layer at same location as source gas file
            if self.layerType == 'GEOPACKAGE' :
                dest = os.path.normpath(os.path.join(os.path.dirname(src), os.path.splitext(src)[0] + ".gpkg"))
                QgsMessageLog.logMessage(f"V {buildID} Creating Geopackage Layer. File dest : {dest}", 'ioGAS', level=Qgis.Info)
                layer = app.createGeoPackageLayer(dest,layerName)


            # Option 2 : Create as a Memory-Only Layer
            if self.layerType == 'SCRATCH':
                layerName = layerName + " [Scratch]"
                QgsMessageLog.logMessage(f"V {buildID} Creating Memory Layer.", 'ioGAS', level=Qgis.Info)
                layer = app.createMemoryLayer(layerName)


            # add layer to the map
            if layer.isValid():
                QgsProject.instance().addMapLayer(layer)
                self.iface.messageBar().pushSuccess('ioGAS', src + ' Loaded OK. ' + buildID)
            else:
                self.iface.messageBar().pushWarning('ioGAS', src + ' Failed to load file')


