import unittest
from gasfileloader import GasFileLoader

GASFILE = "testdata\\gis.small.gas"
GASFILE_631 = "testdata\\gis.small.6.3.1.gas"
GASFILE_61 = "testdata\\gis.small.6.1.gas"
GASFILE_NOSEL = "testdata\\gis.small.nosel.gas"
GASFILE_NONORTH = "testdata\\gis.small.no.north.gas"
GASFILE_DUPEAST = "testdata\\gis.small.dupe.east.gas"


class TestGasFileLoader(unittest.TestCase):



    def test_atts(self):
        g = GasFileLoader(GASFILE)

        #colours
        self.assertEqual(6, len(g.colours)) #6  colours in test file
        self.assertEqual("Default Colour", g.colours[0].name)
        self.assertEqual("A", g.colours[1].name) #all properties of first one
        self.assertEqual("-65536", g.colours[1].colour) #red
        self.assertTrue(g.colours[1].visible)
        self.assertEqual("-16711936", g.colours[2].colour) #2nd green
        self.assertFalse(g.colours[4].visible)
        self.assertEqual("INVIZ", g.colours[4].name)
        self.assertEqual("D", g.colours[5].name)

        # shapes
        self.assertEqual(3, len(g.shapes))
        self.assertEqual("aa", g.shapes[1].name)
        self.assertEqual("1", g.shapes[1].shapeCode)
        self.assertTrue(g.shapes[1].visible)
        self.assertEqual("2", g.shapes[2].shapeCode)

        # sizes
        self.assertEqual(4, len(g.sizes))
        self.assertEqual("_x", g.sizes[1].name)
        self.assertEqual("8", g.sizes[1].size)
        self.assertTrue(g.sizes[1].visible)
        self.assertEqual("10", g.sizes[2].size)

        # titles
        self.assertEqual("atta, attb, _attc", g.symbologyTitle)

        #unique att combinations - sorted
        self.assertEqual(7, len(g.uniqueAttCombinations))
        self.assertEqual((3,1,1,1), (g.uniqueAttCombinations[0]))
        self.assertEqual((3,1,1,2), (g.uniqueAttCombinations[1]))
        self.assertEqual((3,2,2,2), (g.uniqueAttCombinations[2]))


    def test_data(self):
        g = GasFileLoader(GASFILE)
        d = g.data
        a = g.atts
        da = g.dataAndAtts

        #special and selected column names
        #todo put in test below: self.assertEqual(('id', 'east', 'north', 'RL', 'copper', 'tin', 'lead', 'code'),  g.columnNames)

        #column names and types
        self.assertEqual("id", str(g.columns[0].name))
        self.assertEqual("Numeric", str(g.columns[0].type))
        self.assertEqual("copper_", str(g.columns[4].name)) #deliberate _ in name to test escaping
        self.assertEqual("Numeric", str(g.columns[4].type))
        self.assertEqual("code", str(g.columns[7].name))
        self.assertEqual("Text", str(g.columns[7].type))

        #visbile row count
        self.assertEqual(9, len(d))

        #4 sel and 4 special cols
        self.assertEqual(8, len(d[0]))

        #4 att columns
        self.assertEqual(4, len(a[0]))

        #both
        self.assertEqual(4+8, len(da[0]))

        #data values for copper
        self.assertEqual("4", d[0][4])
        self.assertEqual("", d[1][4]) #null

        #data values for code
        self.assertEqual("hello", d[0][7])
        self.assertEqual("there_1", d[1][7])
        self.assertEqual("", d[2][7]) #null

        #test colour of 4th row
        colourAtt = int(a[3][1])
        colour = g.colours[colourAtt]
        self.assertEqual("D", colour.name)
        self.assertEqual("-16776961", colour.colour) #blue

    def test_noSelection(self):
        g = GasFileLoader(GASFILE_NOSEL)
        d = g.data

        # visbile row count
        self.assertEqual(9, len(d))

        # 0 sel and 4 special cols
        self.assertEqual(4, len(d[0]))

    def test_noNorth(self):
        try:
            g = GasFileLoader(GASFILE_NONORTH)
            d = g.data
            self.fail()
        except ValueError as e:
            self.assertEqual('East and North fields must be set in ioGAS prior to saving', str(e))

    def test_dupEast(self):

        g = GasFileLoader(GASFILE_DUPEAST)
        d = g.data
        a = g.atts
        da = g.dataAndAtts

        # column names and types
        print (g.columns)
        self.assertEqual("id", str(g.columns[0].name))
        self.assertEqual("Numeric", str(g.columns[0].type))
        self.assertEqual("east", str(g.columns[1].name))
        self.assertEqual("Numeric", str(g.columns[1].type))
        self.assertEqual("east1", str(g.columns[2].name))  # duplicate east confuses qgis, so we rename in loader
        self.assertEqual("Numeric", str(g.columns[2].type))
        self.assertEqual("east11", str(g.columns[3].name)) #they have set 3 special cols to the same one (this won't happen!)
        self.assertEqual("Numeric", str(g.columns[3].type))

        # 4 sel and 4 special cols
        self.assertEqual(8, len(d[0]))

        # data values for copper
        self.assertEqual("4", d[0][4])
        self.assertEqual("", d[1][4])  # null

    def test_versioning(self):
        try:
            #correct version number reading
            g = GasFileLoader(GASFILE_631)
            self.assertEqual(6003, g.version)
            g = GasFileLoader(GASFILE)
            self.assertEqual(999999, g.version) #dev version
        except:
            self.fail()

        try:
            #can't load old versions
            g = GasFileLoader(GASFILE_61)
            self.assertEqual(6001, g.version)
            self.fail()

        except ValueError as e:
            self.assertEqual("File must have been saved with ioGAS version 6.3 or later", str(e))




    def test_specialColNames(self):
        g = GasFileLoader(GASFILE)

        self.assertEqual("east", g.specialColumnEast)
        self.assertEqual("north", g.specialColumnNorth)
        self.assertEqual("32613", g.epsg)
        self.assertEqual("id", g.specialColumnID)

        self.assertEqual("id", g.columns[0].name)
        #east is *2nd* (id/e/n/elev are at start) despite being selected at the *end*
        self.assertEqual("east", g.columns[1].name)


if __name__ == '__main__':
    unittest.main()